@@ -1,8 +1,25 @@
+#  Licensed to Elasticsearch B.V. under one or more contributor
+#  license agreements. See the NOTICE file distributed with
+#  this work for additional information regarding copyright
+#  ownership. Elasticsearch B.V. licenses this file to you under
+#  the Apache License, Version 2.0 (the "License"); you may
+#  not use this file except in compliance with the License.
+#  You may obtain a copy of the License at
+#
+# 	http://www.apache.org/licenses/LICENSE-2.0
+#
+#  Unless required by applicable law or agreed to in writing,
+#  software distributed under the License is distributed on an
+#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
+#  KIND, either express or implied.  See the License for the
+#  specific language governing permissions and limitations
+#  under the License.
+
 from operator import methodcaller
 import time
 
 from ..exceptions import TransportError
-from ..compat import map, string_types, Queue
+from ..compat import map, string_types, Queue, Mapping
 
 from .errors import ScanError, BulkIndexError
 
@@ -26,22 +43,52 @@ def expand_action(data):
     data = data.copy()
     op_type = data.pop("_op_type", "index")
     action = {op_type: {}}
+
+    # If '_source' is a dict use it for source
+    # otherwise if op_type == 'update' then
+    # '_source' should be in the metadata.
+    if (
+        op_type == "update"
+        and "_source" in data
+        and not isinstance(data["_source"], Mapping)
+    ):
+        action[op_type]["_source"] = data.pop("_source")
+
     for key in (
+        "_id",
         "_index",
+        "_if_seq_no",
+        "_if_primary_term",
         "_parent",
         "_percolate",
+        "_retry_on_conflict",
         "_routing",
         "_timestamp",
-        "routing",
         "_type",
         "_version",
         "_version_type",
-        "_id",
-        "retry_on_conflict",
+        "if_seq_no",
+        "if_primary_term",
+        "parent",
         "pipeline",
+        "retry_on_conflict",
+        "routing",
+        "version",
+        "version_type",
     ):
         if key in data:
-            action[op_type][key] = data.pop(key)
+            if key in {
+                "_if_seq_no",
+                "_if_primary_term",
+                "_parent",
+                "_retry_on_conflict",
+                "_routing",
+                "_version",
+                "_version_type",
+            }:
+                action[op_type][key[1:]] = data.pop(key)
+            else:
+                action[op_type][key] = data.pop(key)
 
     # no data payload for delete
     if op_type == "delete":
@@ -50,91 +97,77 @@ def expand_action(data):
     return action, data.get("_source", data)
 
 
+class _ActionChunker:
+    def __init__(self, chunk_size, max_chunk_bytes, serializer):
+        self.chunk_size = chunk_size
+        self.max_chunk_bytes = max_chunk_bytes
+        self.serializer = serializer
+
+        self.size = 0
+        self.action_count = 0
+        self.bulk_actions = []
+        self.bulk_data = []
+
+    def feed(self, action, data):
+        ret = None
+        raw_data, raw_action = data, action
+        action = self.serializer.dumps(action)
+        # +1 to account for the trailing new line character
+        cur_size = len(action.encode("utf-8")) + 1
+
+        if data is not None:
+            data = self.serializer.dumps(data)
+            cur_size += len(data.encode("utf-8")) + 1
+
+        # full chunk, send it and start a new one
+        if self.bulk_actions and (
+            self.size + cur_size > self.max_chunk_bytes
+            or self.action_count == self.chunk_size
+        ):
+            ret = (self.bulk_data, self.bulk_actions)
+            self.bulk_actions, self.bulk_data = [], []
+            self.size, self.action_count = 0, 0
+
+        self.bulk_actions.append(action)
+        if data is not None:
+            self.bulk_actions.append(data)
+            self.bulk_data.append((raw_action, raw_data))
+        else:
+            self.bulk_data.append((raw_action,))
+
+        self.size += cur_size
+        self.action_count += 1
+        return ret
+
+    def flush(self):
+        ret = None
+        if self.bulk_actions:
+            ret = (self.bulk_data, self.bulk_actions)
+            self.bulk_actions, self.bulk_data = [], []
+        return ret
+
+
 def _chunk_actions(actions, chunk_size, max_chunk_bytes, serializer):
     """
     Split actions into chunks by number or size, serialize them into strings in
     the process.
     """
-    bulk_actions, bulk_data = [], []
-    size, action_count = 0, 0
+    chunker = _ActionChunker(
+        chunk_size=chunk_size, max_chunk_bytes=max_chunk_bytes, serializer=serializer
+    )
     for action, data in actions:
-        raw_data, raw_action = data, action
-        action = serializer.dumps(action)
-        # +1 to account for the trailing new line character
-        cur_size = len(action.encode("utf-8")) + 1
-
-        if data is not None:
-            data = serializer.dumps(data)
-            cur_size += len(data.encode("utf-8")) + 1
-
-        # full chunk, send it and start a new one
-        if bulk_actions and (
-            size + cur_size > max_chunk_bytes or action_count == chunk_size
-        ):
-            yield bulk_data, bulk_actions
-            bulk_actions, bulk_data = [], []
-            size, action_count = 0, 0
-
-        bulk_actions.append(action)
-        if data is not None:
-            bulk_actions.append(data)
-            bulk_data.append((raw_action, raw_data))
-        else:
-            bulk_data.append((raw_action,))
-
-        size += cur_size
-        action_count += 1
-
-    if bulk_actions:
-        yield bulk_data, bulk_actions
+        ret = chunker.feed(action, data)
+        if ret:
+            yield ret
+    ret = chunker.flush()
+    if ret:
+        yield ret
 
 
-def _process_bulk_chunk(
-    client,
-    bulk_actions,
-    bulk_data,
-    raise_on_exception=True,
-    raise_on_error=True,
-    *args,
-    **kwargs
-):
-    """
-    Send a bulk request to elasticsearch and process the output.
-    """
+def _process_bulk_chunk_success(resp, bulk_data, raise_on_error=True):
     # if raise on error is set, we need to collect errors per chunk before raising them
     errors = []
 
-    try:
-        # send the actual request
-        resp = client.bulk("\n".join(bulk_actions) + "\n", *args, **kwargs)
-    except TransportError as e:
-        # default behavior - just propagate exception
-        if raise_on_exception:
-            raise e
-
-        # if we are not propagating, mark all actions in current chunk as failed
-        err_message = str(e)
-        exc_errors = []
-
-        for data in bulk_data:
-            # collect all the information about failed actions
-            op_type, action = data[0].copy().popitem()
-            info = {"error": err_message, "status": e.status_code, "exception": e}
-            if op_type != "delete":
-                info["data"] = data[1]
-            info.update(action)
-            exc_errors.append({op_type: info})
-
-        # emulate standard behavior for failed actions
-        if raise_on_error:
-            raise BulkIndexError(
-                "%i document(s) failed to index." % len(exc_errors), exc_errors
-            )
-        else:
-            for err in exc_errors:
-                yield False, err
-            return
-
     # go through request-response pairs and detect failures
     for data, (op_type, item) in zip(
         bulk_data, map(methodcaller("popitem"), resp["items"])
@@ -155,6 +188,66 @@ def _process_bulk_chunk(
         raise BulkIndexError("%i document(s) failed to index." % len(errors), errors)
 
 
+def _process_bulk_chunk_error(
+    error, bulk_data, raise_on_exception=True, raise_on_error=True
+):
+    # default behavior - just propagate exception
+    if raise_on_exception:
+        raise error
+
+    # if we are not propagating, mark all actions in current chunk as failed
+    err_message = str(error)
+    exc_errors = []
+
+    for data in bulk_data:
+        # collect all the information about failed actions
+        op_type, action = data[0].copy().popitem()
+        info = {"error": err_message, "status": error.status_code, "exception": error}
+        if op_type != "delete":
+            info["data"] = data[1]
+        info.update(action)
+        exc_errors.append({op_type: info})
+
+    # emulate standard behavior for failed actions
+    if raise_on_error:
+        raise BulkIndexError(
+            "%i document(s) failed to index." % len(exc_errors), exc_errors
+        )
+    else:
+        for err in exc_errors:
+            yield False, err
+
+
+def _process_bulk_chunk(
+    client,
+    bulk_actions,
+    bulk_data,
+    raise_on_exception=True,
+    raise_on_error=True,
+    *args,
+    **kwargs
+):
+    """
+    Send a bulk request to elasticsearch and process the output.
+    """
+    try:
+        # send the actual request
+        resp = client.bulk("\n".join(bulk_actions) + "\n", *args, **kwargs)
+    except TransportError as e:
+        gen = _process_bulk_chunk_error(
+            error=e,
+            bulk_data=bulk_data,
+            raise_on_exception=raise_on_exception,
+            raise_on_error=raise_on_error,
+        )
+    else:
+        gen = _process_bulk_chunk_success(
+            resp=resp, bulk_data=bulk_data, raise_on_error=raise_on_error
+        )
+    for item in gen:
+        yield item
+
+
 def streaming_bulk(
     client,
     actions,
@@ -343,8 +436,10 @@ def parallel_bulk(
 
     class BlockingPool(ThreadPool):
         def _setup_queues(self):
-            super(BlockingPool, self)._setup_queues()
-            self._inqueue = Queue(queue_size)
+            super(BlockingPool, self)._setup_queues()  # type: ignore
+            # The queue must be at least the size of the number of threads to
+            # prevent hanging when inserting sentinel values during teardown.
+            self._inqueue = Queue(max(queue_size, thread_count))
             self._quick_put = self._inqueue.put
 
     pool = BlockingPool(thread_count)
@@ -424,50 +519,43 @@ def scan(
     if not preserve_order:
         query = query.copy() if query else {}
         query["sort"] = "_doc"
+
     # initial search
     resp = client.search(
         body=query, scroll=scroll, size=size, request_timeout=request_timeout, **kwargs
     )
-
     scroll_id = resp.get("_scroll_id")
-    if scroll_id is None:
-        return
 
     try:
-        first_run = True
-        while True:
-            # if we didn't set search_type to scan initial search contains data
-            if first_run:
-                first_run = False
-            else:
-                resp = client.scroll(
-                    scroll_id=scroll_id,
-                    scroll=scroll,
-                    request_timeout=request_timeout,
-                    **scroll_kwargs
-                )
-
+        while scroll_id and resp["hits"]["hits"]:
             for hit in resp["hits"]["hits"]:
                 yield hit
 
-            # check if we have any errrors
-            if resp["_shards"]["successful"] < resp["_shards"]["total"]:
+            # check if we have any errors
+            if (resp["_shards"]["successful"] + resp["_shards"]["skipped"]) < resp[
+                "_shards"
+            ]["total"]:
                 logger.warning(
-                    "Scroll request has only succeeded on %d shards out of %d.",
+                    "Scroll request has only succeeded on %d (+%d skipped) shards out of %d.",
                     resp["_shards"]["successful"],
+                    resp["_shards"]["skipped"],
                     resp["_shards"]["total"],
                 )
                 if raise_on_error:
                     raise ScanError(
                         scroll_id,
-                        "Scroll request has only succeeded on %d shards out of %d."
-                        % (resp["_shards"]["successful"], resp["_shards"]["total"]),
+                        "Scroll request has only succeeded on %d (+%d skipped) shards out of %d."
+                        % (
+                            resp["_shards"]["successful"],
+                            resp["_shards"]["skipped"],
+                            resp["_shards"]["total"],
+                        ),
                     )
-
+            resp = client.scroll(
+                body={"scroll_id": scroll_id, "scroll": scroll}, **scroll_kwargs
+            )
             scroll_id = resp.get("_scroll_id")
-            # end of scroll
-            if scroll_id is None or not resp["hits"]["hits"]:
-                break
+
     finally:
         if scroll_id and clear_scroll:
             client.clear_scroll(body={"scroll_id": [scroll_id]}, ignore=(404,))
@@ -516,7 +604,6 @@ def reindex(
         :func:`~elasticsearch.helpers.bulk`
     """
     target_client = client if target_client is None else target_client
-
     docs = scan(client, query=query, index=source_index, scroll=scroll, **scan_kwargs)
 
     def _change_doc_index(hits, index):
